@@ -1,6 +1,6 @@
 ;;; Dezyne --- Dezyne command line tools
 ;;;
-;;; Copyright © 2017, 2018, 2019, 2020, 2022, 2022 Jan (janneke) Nieuwenhuizen <janneke@gnu.org>
+;;; Copyright © 2017, 2018, 2019, 2020, 2022, 2023 Janneke Nieuwenhuizen <janneke@gnu.org>
 ;;; Copyright © 2017 Paul Hoogendijk <paul@dezyne.org>
 ;;; Copyright © 2017, 2018, 2020 Rutger van Beusekom <rutger@dezyne.org>
 ;;; Copyright © 2017 Rob Wieringa <rma.wieringa@gmail.com>
@@ -31,7 +31,7 @@
   #:use-module ((oop goops)
                 #:select (class-slots slot-definition-name slot-ref))
 
-  #:use-module (dzn ast goops)
+  #:use-module (dzn ast ast)
   #:use-module (dzn ast util)
   #:use-module (dzn command-line)
   #:export (ast:display
@@ -41,7 +41,7 @@
 (define (ast port) (display #\* port))
 (define (ref port) (display #\@ port))
 
-(define-method (sdisplay (o <ast-node>) port)
+(define-method (sdisplay (o <ast>) port)
   (display #\space port)
   (write o port))
 
@@ -49,7 +49,7 @@
   (display #\space port)
   (write o port))
 
-(define-method (sdisplay (o <location-node>) port)
+(define-method (sdisplay (o <location>) port)
   (when (%locations?)
     (next-method)))
 
@@ -72,27 +72,21 @@
                (else (sdisplay value port))))))
    (class-slots (class-of o))))
 
-(define-method (write (o <ast-node>) port)
+(define-method (write (o <ast>) port)
   (display "(" port)
   (display (ast-name o) port)
   (ast port)
   (display-slots o port)
   (display #\) port))
 
-(define-method (write (o <ast>) port)
-  (display "(" port)
-  (display (ast-name o) port)
-  (ast port)
-  (ast port)
-  (if (.node o) (display-slots (.node o) port))
-  (display #\) port))
-
 (define ast:display write)
 
-(define* (ast:pretty-print ast #:optional (port (current-output-port)))
+(define* (ast:pretty-print ast #:optional (port (current-output-port))
+                           #:key (width 79))
   "Turn @var{ast} into a user-friendly list representation and
 pretty-print it to PORT."
   (pretty-print (with-input-from-string
                     (with-output-to-string (cut write ast))
                   read)
-                port))
+                port
+                #:width width))
