@@ -1,6 +1,6 @@
 ;;; Dezyne --- Dezyne command line tools
 ;;;
-;;; Copyright © 2014, 2015, 2016, 2017, 2018, 2019, 2020, 2021, 2022, 2023 Jan (janneke) Nieuwenhuizen <janneke@gnu.org>
+;;; Copyright © 2014, 2015, 2016, 2017, 2018, 2019, 2020, 2021, 2022, 2023 Janneke Nieuwenhuizen <janneke@gnu.org>
 ;;; Copyright © 2014, 2018, 2020, 2021, 2022, 2023, 2025, 2026 Rutger van Beusekom <rutger@dezyne.org>
 ;;; Copyright © 2017, 2018, 2019, 2020 Rob Wieringa <rma.wieringa@gmail.com>
 ;;; Copyright © 2017, 2018, 2020 Johri van Eerd <vaneerd.johri@gmail.com>
@@ -166,7 +166,7 @@
     (if (not compound) '()
         (let* ((statements (ast:statement* compound))
                (path (ast:path o)))
-          (take-while (negate (cute member <> path ast:eq?)) statements)))))
+          (take-while (negate (cute memq <> path)) statements)))))
 
 
 ;;;
@@ -186,11 +186,11 @@
        (string->symbol o))
       (((and (? string?) o) ..1)        ;scope
        (string->symbol (string-join o ".")))
-      (($ <scope.name-node>)            ;name
+      (($ <scope.name>)                 ;name
        (object->key (string-join (.ids o) ".")))
       (_ o)))                           ;eq?-ness already stable.
   (lambda args
-    (let* ((key (cons base-func (map (compose object->key ast:unwrap) args)))
+    (let* ((key (cons base-func (map object->key args)))
            (cached (hashx-ref funcq-hash funcq-assoc funcq-memo key not-found)))
       (funcq-buffer key)
       (if (not (eq? cached not-found)) cached
@@ -223,7 +223,6 @@
                         (filter (match-lambda ("/" #f) (o o)) scope)))
              (global? (and (pair? scope) (equal? "/" (car scope))))
              (target (if (null? scope) name (car scope)))
-             (foo (ast:declaration* context))
              (found (filter (cute ast:name-equal? <> target)
                             (ast:declaration* context))))
         (and (pair? found)
@@ -283,10 +282,10 @@ null) and return its CONTEXT."
       (search-or-widen-context scope name context)))))
 
 (define ast:lookup-memoized (ast:perfect-funcq ast:lookup-unmemoized))
-(define-method (ast:lookup (context <ast>) name)
-  (let* ((context (or (as context <scope>) (ast:parent context <scope>)))
-         (root (or (as context <root>) (ast:parent context <root>))))
-    (ast:lookup-memoized root context name)))
+(define-method (ast:lookup (o <ast>) name)
+  (let ((scope (or (as o <scope>) (ast:parent o <scope>)))
+        (root (or (as o <root>) (ast:parent o <root>))))
+    (ast:lookup-memoized root scope name)))
 (define-method (ast:lookup name)
   (ast:lookup name name))
 
@@ -319,7 +318,7 @@ null) and return its CONTEXT."
      (name? o))
     (($ <formal-binding>)
      (name? o))
-    (($ <on>)
+    ((or ($ <on>) ($ <canonical-on>))
      (or (find (cute ast:lookup-variable <> name statements)
                (append-map ast:formal* (ast:trigger* o)))
          (ast:lookup-variable (.parent o) name statements)))
@@ -393,10 +392,10 @@ null) and return its CONTEXT."
                         (ast:parent o <interface>))))
     (cond ((and (not port-name)
                 (equal? (.event.name o) "inevitable"))
-           (clone (ast:inevitable) #:parent interface))
+           (graft interface (make <inevitable>)))
           ((and (not port-name)
                 (equal? (.event.name o) "optional"))
-           (clone (ast:optional) #:parent interface))
+           (graft interface (make <optional>)))
           (else (and
                  interface
                  (let ((event (ast:lookup interface (.event.name o))))
@@ -408,11 +407,10 @@ null) and return its CONTEXT."
 (define-method (.event.direction (o <trigger>))
   ((compose .direction .event) o))
 
-(define-method (.function (model <model>) (o <call>))
-  (and (.function.name o) (ast:lookup model (.function.name o))))
-
 (define-method (.function (o <call>))
-  (and (.function.name o) (ast:lookup o (.function.name o))))
+  (and (.function.name o)
+       (and=> (ast:lookup o (.function.name o))
+              (cute as <> <function>))))
 
 (define-method (.variable (o <assign>))
   (and=> (.variable.name o) (cute ast:lookup-variable o <>)))
@@ -467,7 +465,8 @@ null) and return its CONTEXT."
 (define-method (ast:event-formal (o <formal>))
   (let* ((trigger (ast:parent o <trigger>))
          (event (.event trigger))
-         (index (list-index (cute ast:eq? o <>) (reverse (.elements (.parent o)))))
+         (trigger-formals (ast:formal* (.parent o)))
+         (index (list-index (cute eq? o <>) (reverse trigger-formals)))
          (formals (if (not event) '() (ast:formal* event))))
     (and event (< index (length formals)) (list-ref (reverse formals) index))))
 
@@ -475,7 +474,8 @@ null) and return its CONTEXT."
   (let* ((on (ast:parent o <on>))
          (trigger (car (ast:trigger* on)))
          (event (.event trigger))
-         (index (list-index (cute ast:eq? o <>) (reverse (.elements (.parent o)))))
+         (trigger-formals (ast:formal* (.parent o)))
+         (index (list-index (cute eq? o <>) (reverse trigger-formals)))
          (formals (ast:formal* event)))
     (and event (< index (length formals)) (list-ref (reverse formals) index))))
 
@@ -487,11 +487,17 @@ null) and return its CONTEXT."
                     (and=> (ast:parent o <scope>) .parent))))
     (if type-name (ast:lookup scope type-name)
         (let ((formal (ast:event-formal o)))
-          (and formal (.type formal))))))
+          (and=> formal .type)))))
 
 (define-method (.direction (o <formal>))
   (let ((type-name (.type.name o)))
-    (if type-name (.direction (.node o))
+    (if type-name ((@ (oop goops) slot-ref) o 'direction)
+        (let ((formal (ast:event-formal o)))
+          (and formal (.direction formal))))))
+
+(define-method (.direction (o <formal-binding>))
+  (let ((type-name (.type.name o)))
+    (if type-name ((@ (oop goops) slot-ref) o 'direction)
         (let ((formal (ast:event-formal o)))
           (and formal (.direction formal))))))
 
