@@ -51,14 +51,14 @@
 
   #:declarative? #f
 
-  #:export (%model-name
+  #:export (%model
             makreel:.name
             makreel:arguments
             makreel:call-continuation*
-            makreel:called-function*
             makreel:constraint
             makreel:defer*
             makreel:defer-skip?
+            makreel:direct->direct+indirect
             makreel:enum-fields
             makreel:full-name
             makreel:get-model
@@ -77,10 +77,11 @@
             makreel:proc-list
             makreel:process-identifier
             makreel:variables-in-scope
+            model-name
             root->))
 
 (define %id-alist (make-parameter #f))
-(define %model-name (make-parameter #f))
+(define %model (make-parameter #f))
 (define %next-alist (make-parameter #f))
 
 
@@ -104,10 +105,10 @@
 (define (makreel:model->makreel root model)
   (let* ((model-name (ast:dotted-name model))
          (root' (tree-filter (disjoin (negate (is? <component>))
-                                      (cut ast:eq? <> model))
+                                      (cute eq? <> model))
                              root)))
     (parameterize ((%language "makreel")
-                   (%model-name model-name))
+                   (%model model))
       (root-> root'))))
 
 (define (makreel:unticked-dotted-name o)
@@ -125,7 +126,7 @@
 (define-method (makreel:defer*-unmemoized (o <ast>))
   (tree-collect (conjoin (is? <defer>)
                          (disjoin (negate (cute ast:parent <> <function>))
-                                  (compose is-called? (cute ast:parent <> <function>))))
+                                  (cute ast:parent <> <function>)))
                 o))
 
 (define makreel:defer*
@@ -148,42 +149,6 @@
 (define-method (makreel:full-name (o <shared-variable>))
   (string-append (.port.name o) "port_" (.name o)))
 
-(define (reachable calls)
-  (let ((nested direct (partition (cute ast:parent <> <function>) calls)))
-    (let loop ((nested nested) (direct direct))
-      (let ((reached rest (partition
-                           (lambda (call)
-                             (find (cut ast:eq? (ast:parent call <function>) <>)
-                                   (map .function direct))) nested)))
-        (let* ((reached (append reached direct)))
-          (if (equal? direct reached) direct
-              (loop rest reached)))))))
-
-(define (reachable-calls-unmemoized root o)
-  (let* ((calls (tree-collect-filter
-                 (disjoin (is? <behavior>)
-                          (is? <declarative>)
-                          (is? <functions>)
-                          (is? <function>)
-                          (is? <statement>))
-                 (conjoin (is? <call>) ast:imperative?)
-                 o))
-         (calls (reachable calls)))
-    calls))
-
-(define (reachable-calls o)
-  ((ast:perfect-funcq reachable-calls-unmemoized) (ast:parent o <root>) o))
-
-(define-method (is-called? (o <function>))
-  (let ((calls (reachable-calls (ast:parent o <behavior>))))
-    (find (compose (cut equal? <> (.name o)) .function.name) calls)))
-
-(define-method (makreel:called-function* (o <behavior>))
-  (filter is-called? (ast:function* o)))
-
-(define-method (makreel:called-function* (o <model>))
-  ((compose makreel:called-function* .behavior) o))
-
 (define-method (makreel:process-index (o <statement>))
   (makreel:process-identifier o))
 
@@ -201,7 +166,7 @@
   (if (is-a? o <behavior>) '()
       (let* ((p (.parent o)))
         (cond ((is-a? p <compound>)
-               (let ((pre (cdr (member o (reverse (ast:statement* p)) ast:eq?))))
+               (let ((pre (cdr (memq o (reverse (ast:statement* p))))))
                  (append (filter (is? <variable>) pre) (makreel:locals p))))
               ((is-a? p <defer>)
                (let ((model (ast:parent p <model>)))
@@ -212,7 +177,8 @@
                (makreel:locals p))))))
 
 (define-method (makreel:member* (o <ast>))
-  (ast:member* (ast:parent o <model>)))
+  (let ((model (ast:parent o <model>)))
+    (if model (ast:member* model) '())))
 
 (define-method (makreel:variables-in-scope (o <model>))
   (ast:member* o))
@@ -221,7 +187,7 @@
   (let ((stack (and (or (as o <function>)
                         (ast:parent o <function>))
                     (not (ast:parent o <defer>))
-                    (clone (make <stack>) #:parent o))))
+                    (graft o (make <stack>)))))
     (append (makreel:member* o)
             (makreel:locals o)
             (if (not stack) '()
@@ -253,25 +219,32 @@
     (#f '())))
 
 (define-method (makreel:enum-fields (o <enum>))
-  (map (compose (cut clone <> #:parent o)
-                (cut make <enum-literal> #:type.name (.name o) #:field <>))
+  (map (compose (cute graft o <>)
+                (cute make <enum-literal> #:type.name (.name o) #:field <>))
        (ast:field* o)))
 
+(define (makreel:direct->direct+indirect direct-calls)
+  "Given that every call in a behavior points to a called function and
+every called function may call zero or more functions.  Return the list
+of all unique called functions and calls."
+  (let loop ((calls direct-calls) (functions '()) (direct+indirect-calls '()))
+    (if (null? calls) (values direct+indirect-calls functions)
+        (let* ((call (car calls))
+               (function (.function call)))
+          (if (find (cute eq? function <>) functions)
+              (loop (cdr calls) functions (cons call direct+indirect-calls))
+              (loop (append (cdr calls)
+                            (tree-collect (is? <call>) function))
+                    (lset-adjoin eq? functions function)
+                    (cons call direct+indirect-calls)))))))
+
 (define-method (makreel:call-continuation*-unmemoized (o <behavior>))
-  (let* ((calls (tree-collect (disjoin
-                               (conjoin (is? <call>) ast:imperative?)
-                               (conjoin
-                                (disjoin (is? <variable>)
-                                         (is? <assign>))
-                                (compose (is? <call>)
-                                         .expression)))
-                              o))
-         (called (makreel:called-function* o))
-         (calls (filter (compose
-                         (disjoin not
-                                  (cute member <> called ast:eq?))
-                         (cute ast:parent <> <function>))
-                        calls)))
+  (let* ((calls (tree-collect
+                 (disjoin (conjoin (is? <call>) ast:imperative?)
+                          (conjoin (disjoin (is? <variable>) (is? <assign>))
+                                   (compose (is? <call>) .expression)))
+                 o))
+         (calls (makreel:direct->direct+indirect calls)))
     (cons (map (compose car ast:continuation*) calls)
           calls)))
 
@@ -285,7 +258,8 @@
     (makreel:call-continuation* behavior)))
 
 (define-method (makreel:call-continuation* (o <ast>))
-  (let ((behavior (ast:parent o <behavior>)))
+  (let ((behavior (or (ast:parent o <behavior>)
+                      (.behavior (%model)))))
     (makreel:call-continuation* behavior)))
 
 (define-method (makreel:stack-empty? (o <call>))
@@ -298,14 +272,9 @@
 ;;; Helpers.
 ;;;
 (define-method (makreel:init (o <root>))
-  (let* ((model-name (%model-name)))
-    (define (named? o)
-      (equal? (makreel:unticked-dotted-name o) model-name))
-    (let ((model (and model-name
-                      (find named? (ast:model** o)))))
-      (or model
-          (find (is? <component>) (ast:model** o))
-          (find (is? <interface>) (ast:model** o))))))
+  (or (%model)
+      (find (is? <component>) (ast:model** o))
+      (find (is? <interface>) (ast:model** o))))
 
 (define (makreel:init-process process)
   (format #f "init ~a;\n" process))
@@ -314,18 +283,20 @@
   (let ((location (.location o)))
     (format #f "~a, ~a" (.line location) (.column location))))
 
-(define (makreel:process-identifier o)
-  (let* ((model (ast:parent o <model>))
+(define* (makreel:process-identifier o #:key model)
+  (let* ((model (or (ast:parent o <model>) model))
+         ;; HACK; FIXME
          (model-key (.id model))
          (path (ast:path o))
          (key (map .id path))
          (next (assq-ref (%next-alist) model-key))
-         (next (or next -1)))
-    (number->string (or (assoc-ref (%id-alist) key)
-                        (let ((next (1+ next)))
-                          (%id-alist (acons key next (%id-alist)))
-                          (%next-alist (assoc-set! (%next-alist) model-key next))
-                          next)))))
+         (next (or next -1))
+         (next (or (assoc-ref (%id-alist) key)
+                   (let ((next (1+ next)))
+                     (%id-alist (acons key next (%id-alist)))
+                     (%next-alist (assoc-set! (%next-alist) model-key next))
+                     next))))
+    (number->string next)))
 
 (define-method (makreel:process-parens (o <ast>))
   (and (or (pair? (makreel:variables-in-scope o))
@@ -430,7 +401,7 @@
               #:statement ((compose (tick-names- names) .statement) o))))
     (($ <call>)
      (clone o
-            #:function.name ((compose (append-tick names) .function.name) o)
+            #:function.name ((compose (tick-names- names) .function.name) o)
             #:arguments ((compose (tick-names- names) .arguments) o)))
     (($ <assign>)
      (clone o
@@ -497,7 +468,8 @@
       (((and statement ($ <action>)
              (? ast:requires?) (? ast:blocking?))
         rest ...)
-       (let ((action-reply (make <action-reply> #:action statement)))
+       (let* ((action-reply (make <action-reply> #:action statement))
+              (action-reply (graft (.parent statement) action-reply)))
          (cons* statement action-reply (add-action-reply rest))))
       (((and statement (or ($ <assign>) ($ <variable>))
              (= .expression
@@ -508,7 +480,8 @@
                         (.variable.name statement)))
               (action-reply (make <action-reply>
                               #:action action
-                              #:variable.name name)))
+                              #:variable.name name))
+              (action-reply (graft (.parent action) action-reply)))
          (cons* statement action-reply (add-action-reply rest))))
       ((statement rest ...)
        (cons (add-action-reply statement) (add-action-reply rest)))
@@ -567,9 +540,9 @@
   (match o
     ((? (is? <component>)) o)
     ((? (is? <compound>))
-     (clone o #:elements (makreel:add-state-placeholder (ast:statement* o))))
+     (graft o #:elements (makreel:add-state-placeholder (ast:statement* o))))
     (((and (? (is? <if>)) s) rest ...)
-     (cons (clone s #:expression (.expression s)
+     (cons (graft s #:expression (.expression s)
                   #:then (makreel:add-state-placeholder (.then s))
                   #:else (and=> (.else s) makreel:add-state-placeholder))
            (makreel:add-state-placeholder rest)))
@@ -589,20 +562,20 @@
 transformations."
   (parameterize ((%normalize:short-circuit? makreel:short-circuit?))
     (let ((root ((compose
-                  makreel:add-action-reply
-                  makreel:add-shared-variables
-                  makreel:mark-tail-call
-                  add-function-return
-                  extract-call
-                  makreel:add-state-placeholder
-                  (if (%no-unreachable?) identity tag-imperative-blocks)
-                  normalize:state+illegals
-                  remove-otherwise
-                  makreel:tick-names
-                  add-defer-end
-                  purge-data
-                  (add-explicit-temporaries)
-                  inline-expression-functions)
+                  (with-root makreel:add-action-reply)
+                  (with-root makreel:add-shared-variables)
+                  (with-root makreel:mark-tail-call)
+                  (with-root add-function-return)
+                  (with-root extract-call)
+                  (with-root makreel:add-state-placeholder)
+                  (if (%no-unreachable?) identity (with-root tag-imperative-blocks))
+                  (with-root normalize:state+illegals)
+                  (with-root remove-otherwise)
+                  (with-root makreel:tick-names)
+                  (with-root add-defer-end)
+                  (with-root purge-data)
+                  (with-root (add-explicit-temporaries))
+                  (with-root inline-expression-functions))
                  ast)))
       (when (> (dzn:debugity) 1)
         (debug "normalized root:")
@@ -646,7 +619,7 @@ transformations."
             (shared (makreel:shared* behavior))
             (shared (delete-duplicates shared ast:equal?))
             (shared (map shared-var->shared-variable shared)))
-       (clone o #:elements (append (ast:variable* o) shared))))
+       (graft o #:elements (append (ast:variable* o) shared))))
     ((? (is? <ast>)) (tree-map makreel:add-shared-variables o))
     (_ o)))
 
