@@ -61,8 +61,9 @@
   (let* ((name (ast:full-name o))
          (name (if (not (is-a? o <foreign>)) name
                    (cons "SKEL" name)))
-         (name (if (is-a? o <enum>) (cons "ENUM" name)
-                   (append name '("h"))))
+         (name (cond ((is-a? o <enum>) (cons "ENUM" name))
+                     ((is-a? o <function>) (cons "FUNCTION" name))
+                     (else (append name '("h")))))
          (name (string-join name "_"))
          (guard (string-upcase name)))
     `(,(sm:cpp-ifndef* guard)
@@ -174,6 +175,10 @@
             ,@(c:enum->statements o))))
     (c:include-guard o statements)))
 
+(define (c:global-function->header-statements function)
+  (let ((statement (function->function function)))
+    (c:include-guard function (list statement))))
+
 (define-method (c:->formal (o <formal>))
   (let* ((type (code:type-name o))
          (type (if (ast:in? o) type
@@ -248,7 +253,7 @@
     (define (instance->connect instance)
       (let* ((component (.type instance))
              (ports (ast:injected-port* component))
-             (ports (filter (compose (cute ast:eq? <> interface) .type) ports)))
+             (ports (filter (compose (cute eq? <> interface) .type) ports)))
         (map (cute port->connect instance <>) ports)))
     (append-map instance->connect instances)))
 
@@ -301,11 +306,16 @@
                     arguments)))))
 
 (define-method (ast->code (o <call>))
-  (let* ((model (ast:parent o <model>))
-         (model-name (c:type-name model))
-         (name (string-append model-name "_" (.function.name o)))
+  (let* ((function (.function o))
+         (name (ast:name function))
+         (global? (not (ast:parent function <behavior>)))
+         (name (if global? name
+                   (let* ((model (ast:parent o <model>))
+                          (model-name (c:type-name model)))
+                     (string-append model-name "_" name))))
          (arguments (map ast->expression (ast:argument* o)))
-         (arguments (cons "self" arguments)))
+         (arguments (if global? arguments
+                        (cons "self" arguments))))
     (sm:call
      (name name)
      (arguments arguments))))
@@ -434,13 +444,22 @@
 ;;;
 ;;; Root.
 ;;;
+(define-method (root->header-global-functions (o <root>))
+  (let ((functions (filter (negate ast:imported?) (ast:function** o))))
+    (append-map c:global-function->header-statements functions)))
+
 (define-method (root->header-statements (o <root>))
   (let ((imports (ast:unique-import* o))
         (enums (c:enum* o)))
     (append
      (map c:file-name->include imports)
      (map (compose (cute string-append <> "\n") .value) (ast:data* o))
-     (append-map c:enum->header-statements enums))))
+     (append-map c:enum->header-statements enums)
+     (root->header-global-functions o))))
+
+(define-method (root->global-functions (o <root>))
+  (let ((functions (filter (negate ast:imported?) (ast:function** o))))
+    (map function->function functions)))
 
 (define-method (root->statements (o <root>))
   (let ((enums (c:enum* o)))
@@ -930,6 +949,13 @@
 (define-method (component-model->statements (o <component-model>))
   ((ast:perfect-funcq component-model->sm:statements-unmemoized) o))
 
+(define (function->function function)
+  (sm:function
+   (name (.name function))
+   (type (code:type-name (ast:type function)))
+   (formals (map c:->formal (ast:formal* function)))
+   (statement (ast->code (.statement function)))))
+
 (define-method (model->header-statements (o <component>))
   (let ((interface-includes (code:interface-include* o)))
     (c:include-guard
@@ -1297,6 +1323,7 @@
                       ,(sm:cpp-system-include* "dzn/locator.h")
                       ,@(root->header-statements o)
                       ,@(append-map model->header-statements models)
+                      ,@(root->global-functions o)
                       ,(code:version-comment))))))
     (display header)))
 
@@ -1320,6 +1347,7 @@
                        (sm:cpp-system-include* "string.h"))
                     ,@(root->statements o)
                     ,@(append-map model->statements models)
+                    ,@(root->global-functions o)
                     ,(code:version-comment))))))
     (display code)))
 
