@@ -160,7 +160,7 @@
          (arguments (ast:argument* o))
          (arguments (map ast->expression arguments formals)))
     (sm:call-method
-     (name (.function.name o))
+     (name (ast:name (.function.name o)))
      (arguments arguments))))
 
 (define-method (ast->code (o <reply>))
@@ -427,16 +427,14 @@
            (formals (ast:formal* trigger))
            (arguments (map cs:->argument formals))
            (out-formals (filter (negate ast:in?) formals))
-           ;; (out-arguments (map .name out-formals))
            (formals (map cs:->formal formals))
-           (on (code:on trigger))
-           (on-triggers (and=> on ast:trigger*))
-           (on-trigger (or (and=> on-triggers car) trigger))
-           (formal-bindings (filter (is? <formal-binding>) (ast:formal* on-trigger)))
-           (out-bindings (make <out-bindings>
-                           #:elements formal-bindings
-                           #:port (.port trigger)))
-           (out-bindings (clone out-bindings #:parent on))
+           (formal-bindings (filter (is? <formal-binding>) (ast:formal* trigger)))
+           (on (ast:parent trigger <on>))
+           (parent (or (and=> on .statement)
+                       o))
+           (out-bindings (graft parent (make <out-bindings>
+                                         #:elements formal-bindings
+                                         #:port (.port trigger))))
            (out-binding (code:out-binding port))
            (this-out-binding (code:member out-binding))
            (type (ast:type trigger))
@@ -499,8 +497,7 @@
               (($ <foreign>)
                (values trigger #f))
               (($ <component>)
-               (let* ((on (code:on trigger))
-                      (trigger (car (ast:trigger* on)))
+               (let* ((on (ast:parent trigger <on>))
                       (statement (.statement on)))
                  (values trigger (ast->code statement))))))
            (formals (ast:formal* trigger))
@@ -637,7 +634,7 @@
          (expression (and (not injected?)
                           (port->expression port))))))
     (define (injected-instance? instance)
-      (member instance injected-instances ast:eq?))
+      (memq instance injected-instances))
     (define (instance->expression instance)
       (let ((instance-name (.name instance))
             (type (code:type-name (.type instance))))
@@ -654,7 +651,7 @@
          (type (code:type-name (.type instance)))
          (name instance-name)
          (expression (and (or (not injected?)
-                              (member instance injected-instances ast:eq?))
+                              (memq instance injected-instances))
                           (instance->expression instance))))))
     (define (instance->assignment instance)
       (let ((name (.name instance)))
@@ -762,7 +759,7 @@
          (expression (and (not injected?)
                           (port->expression port))))))
     (define (injected-instance? instance)
-      (member instance injected-instances ast:eq?))
+      (memq instance injected-instances))
     (define (instance->expression instance)
       (let ((instance-name (.name instance))
             (type (code:type-name (.type instance))))
@@ -779,7 +776,7 @@
          (type (code:type-name (.type instance)))
          (name instance-name)
          (expression (and (or (not injected?)
-                              (member instance injected-instances ast:eq?))
+                              (memq instance injected-instances))
                           (instance->expression instance))))))
     (define (instance->assignment instance)
       (let ((name (.name instance)))
