@@ -170,7 +170,8 @@ SEMICOLON < ';'")
 
 imports <-- import+
 
-definitions <-- (dollars / type / namespace / interface / component / EOF)#+
+definitions <-- (dollars / type / namespace / function / prototype /
+                interface / component / EOF)#+
 
 import <-- IMPORT file-name SEMICOLON#
   file-name <- (!SEMICOLON .)+
@@ -189,7 +190,7 @@ type <- enum / int / extern
   extern <-- EXTERN scoped-name# dollars# SEMICOLON#
 
 namespace <-- NAMESPACE compound-name# BRACE-OPEN# namespace-root BRACE-CLOSE#
-  namespace-root <-- (type / namespace / interface / component / &BRACE-CLOSE)#*
+  namespace-root <-- (type / function / prototype / namespace / interface / component / &BRACE-CLOSE)#*
 
 interface <-- INTERFACE scoped-name#
               BRACE-OPEN# types-and-events# (behavior / &BRACE-CLOSE)#
@@ -232,6 +233,9 @@ behavior <-- BEHAVIOR behavior-compound
          <-- type-name (!is-event name / duplicate-identifier) &(formals BRACE-OPEN)
              enter-frame formals compound# exit-frame
 
+      prototype
+         <-- type-name (!is-event name / duplicate-identifier) formals SEMICOLON#
+
       expression-function
          <-- type-name (!is-event name / duplicate-identifier)
              &PAREN-OPEN formals ASSIGN expression# SEMICOLON#
@@ -275,7 +279,7 @@ imperative-statement
 
   action-or-call <- (action / call) SEMICOLON#
     action <-- is-port DOT name# arguments / is-event !arguments
-    call <-- name arguments
+    call <-- compound-name arguments
       arguments <-- PAREN-OPEN (argument (&PAREN-CLOSE / COMMA#))* PAREN-CLOSE#
         argument <-- expression
 
@@ -319,7 +323,8 @@ plus-min-expression
   <- not-expression (PLUS / MINUS) plus-min-expression# / not-expression
 
 not-expression
-  <- minus / negation / group / dollars / literal / !(is-port / var) enum-literal
+  <- minus / negation / group / dollars / literal
+     / !(is-port / call / var) enum-literal
      / field-test / action / shared-field-test / shared-var / call / var
      / unknown-identifier
 
