@@ -194,6 +194,9 @@ output, and standard error as three values."
 (define (flush? file-name)
   (get-meta-flag file-name 'flush))
 
+(define (lts-test? file-name)
+  (get-meta-flag file-name 'lts-test?))
+
 (define (no-unreachable? file-name)
   (get-meta-flag file-name 'no-unreachable))
 
@@ -239,6 +242,11 @@ output, and standard error as three values."
     (or (get-meta-option file-name 'simulate-flags)
         '())))
 
+(define (dzn-flags file-name)
+  (let ((alist (get-meta file-name)))
+    (or (get-meta-option file-name 'dzn-flags)
+        '())))
+
 
 ;;;
 ;;; Utility.
@@ -269,6 +277,8 @@ output, and standard error as three values."
                     (feature? "double_inevitable")
                     (feature? "flush"))
                 "flush")
+           (feature? "foreign_function")
+           (feature? "global_function")
            (feature? "inject")
            (and (or (feature? "reply_data")
                     (feature? "data_reply"))
@@ -335,6 +345,7 @@ output, and standard error as three values."
   (format #t "** stage: verify\n")
   (let* ((base-name (basename file-name))
          (dzn-name (string-append file-name "/" base-name ".dzn"))
+         (dzn-flags (dzn-flags file-name))
          (includes (cons (string-append file-name "/dzn")
                          (or (and=> (getenv "DZN_INCLUDE_PATH")
                                     (cute string-split <> #\:))
@@ -349,10 +360,10 @@ output, and standard error as three values."
          (queue-size-external (queue-size-external file-name))
          (command
           (if fall-back?
-              `("dzn" "--verbose" "parse" "--fall-back"
+              `("dzn" ,@dzn-flags"--verbose" "parse" "--fall-back"
                 ,@includes
                 ,dzn-name)
-              `("dzn" "--verbose" "verify"
+              `("dzn" ,@dzn-flags "--verbose" "verify"
                 ,@includes
                 "--keep-going"
                 ,@(if (not no-unreachable?) '()
@@ -376,6 +387,7 @@ output, and standard error as three values."
   (and (let* ((base-name (basename file-name))
               (dzn-name (string-append file-name "/" base-name ".dzn"))
               (input "")
+              (dzn-flags (dzn-flags file-name))
               (in-lang (string-append file-name "/" language))
               (includes (cons in-lang
                               (or (and=> (getenv "DZN_INCLUDE_PATH")
@@ -387,7 +399,7 @@ output, and standard error as three values."
               (out-lang (string-append out "/" language))
               (model (or (model? file-name) (file->model base-name)))
               (command
-               `("dzn" "code"
+               `("dzn" ,@dzn-flags "code"
                  ,@includes
                  "--language" ,language
                  "--output" ,out-lang
@@ -429,12 +441,14 @@ output, and standard error as three values."
        (or (code-error? file-name)
            (let* ((traces (find-files file-name ".*trace.*$"))
                   (traces (sort traces trace<)))
-             (and-map (cute run-execute file-name language <>) traces)))))
+             (if (lts-test? file-name) (run-lts-test file-name language)
+                 (and-map (cute run-execute file-name language <>) traces))))))
 
 (define (run-traces file-name)
   (format #t "** stage: traces\n")
   (let* ((base-name (basename file-name))
          (dzn-name (string-append file-name "/" base-name ".dzn"))
+         (dzn-flags (dzn-flags file-name))
          (input "")
          (includes (cons (string-append file-name "/dzn")
                          (or (and=> (getenv "DZN_INCLUDE_PATH")
@@ -455,7 +469,7 @@ output, and standard error as three values."
         (skip? file-name "traces")
         (receive (status stdout stderr)
             (observe
-             `("dzn" "traces"
+             `("dzn" ,@dzn-flags "traces"
                ,@includes
                "--model" ,model
                "--output" ,out-lang
@@ -479,6 +493,7 @@ are weak-bisim equivalent"
   (format #t "** stage: constrained-component\n")
   (let* ((base-name (basename file-name))
          (dzn-name (string-append file-name "/" base-name ".dzn"))
+         (dzn-flags (dzn-flags file-name))
          (input "")
          (includes (cons (string-append file-name "/dzn")
                          (or (and=> (getenv "DZN_INCLUDE_PATH")
@@ -498,7 +513,7 @@ are weak-bisim equivalent"
       (let ((stdout
              status
              (pipeline->string
-              `(("dzn" "parse"
+              `(("dzn" @dzn-flags "parse"
                  ,@includes
                  "--model" ,model
                  "--list-models"
@@ -513,7 +528,7 @@ are weak-bisim equivalent"
                stdout-constrained
                stderr-constrained
                (observe
-                `("dzn" "traces"
+                `("dzn" ,@dzn-flags "traces"
                   "--lts"
                   ,@includes
                   "--model" ,model
@@ -561,6 +576,7 @@ are weak-bisim equivalent"
   (format #t "** stage: constrained-no-compliance\n")
   (let* ((base-name (basename file-name))
          (dzn-name (string-append file-name "/" base-name ".dzn"))
+         (dzn-flags (dzn-flags file-name))
          (input "")
          (includes (cons (string-append file-name "/dzn")
                          (or (and=> (getenv "DZN_INCLUDE_PATH")
@@ -607,7 +623,7 @@ are weak-bisim equivalent"
                stdout-no-compliance
                stderr-no-compliance
                (observe
-                `("dzn" "traces"
+                `("dzn" ,@dzn-flags "traces"
                   "--lts"
                   "--no-non-compliance"
                   ,@includes
@@ -704,11 +720,81 @@ are weak-bisim equivalent"
                                 #f))
                    (zero? status))))))))
 
+(define (run-lts-test file-name language)
+  (format #t "** stage: module-test: ~a\n" language)
+  (let* ((base-name (basename file-name))
+         (dzn-name (string-append file-name "/" base-name ".dzn"))
+         (input "")
+         (out (string-append file-name "/out"))
+         (out-lang (string-append out "/" language))
+         (test (string-append out-lang "/test"))
+         (interpreter (or (assoc-ref %interpreter-alist language) '()))
+         (lts (string-append out "/traces/" base-name ".aut"))
+         (i/o-lts (string-append out "/lts/io-lts.aut"))
+         (i/o-result (string-append out "/lts/io-result.aut"))
+         (model (or (model? file-name) base-name)))
+    (mkdir-p (string-append out "/lts"))
+    (or (error-model? file-name)
+        (feature-skip? file-name language)
+        (skip? file-name
+               language
+               "code" (string-append language ":code")
+               "build" (string-append language ":build")
+               "execute" (string-append language ":execute"))
+        (let* ((simulate-lts-file lts)
+               (simulate-lts (with-input-from-file lts read-string))
+               (foreign (list-matches
+                         "\"([^.\"]*[.][^.\"]*[.][^\"]*)\""
+                         simulate-lts))
+               (foreign (map (cute match:substring <> 1) foreign))
+               (foreign (delete-duplicates foreign))
+               (modeling (list-matches
+                          "\"([^\"]*[.](inevitable|optional))\""
+                          simulate-lts))
+               (modeling (map (cute match:substring <> 1) modeling))
+               (modeling (delete-duplicates modeling))
+               (foreign+modeling (append foreign modeling)))
+          (observe
+           `("ltsconvert"
+             "-eweak-trace"
+             ,(string-append "--tau=" (string-join foreign+modeling ","))
+             ,simulate-lts-file
+             ,lts)
+           "")
+          (receive (status stdout stderr)
+              (observe
+               `("dzn"
+                 "lts"
+                 "--i/o-lts"
+                 "-m" ,model
+                 ,dzn-name
+                 ,lts)
+               "")
+            (with-output-to-file i/o-lts (cute display stdout)))
+          (receive (status stdout stderr)
+              (observe `("dzn" "test" "-r" "50"
+                         ,i/o-lts
+                         ,@interpreter
+                         ,test)
+                       input)
+            (with-output-to-file i/o-result (cute display stdout)))
+          (receive (status stdout stderr)
+              (observe
+               `("ltscompare"
+                 "--quiet"
+                 "-ebisim"
+                 ,i/o-lts
+                 ,i/o-result)
+               "")
+            (and (zero? status)
+                 (string-contains stdout "true")))))))
+
 (define (run-simulate-trace file-name trace)
   (format #t "** stage: simulate-trace ~a\n" trace)
   (let* ((base-name (basename file-name))
          (dzn-name (string-append file-name "/" base-name ".dzn"))
          (baseline (string-append file-name "/baseline/simulate"))
+         (dzn-flags (dzn-flags file-name))
          (input (with-input-from-file trace read-string))
          (includes (cons (string-append file-name "/dzn")
                          (or (and=> (getenv "DZN_INCLUDE_PATH")
@@ -729,7 +815,7 @@ are weak-bisim equivalent"
                              (baseline? "trace")
                              (else "event"))))
     (receive (status stdout stderr)
-        (observe `("dzn" "simulate"
+        (observe `("dzn" ,@dzn-flags "simulate"
                    ,@includes
                    ,@(if (not queue-size) '()
                          `("--queue-size" ,(number->string queue-size)))
@@ -788,6 +874,7 @@ are weak-bisim equivalent"
   (format #t "** stage: lts\n")
   (let* ((base-name (basename file-name))
          (dzn-name (string-append file-name "/" base-name ".dzn"))
+         (dzn-flags (dzn-flags file-name))
          (input "")
          (includes (cons (string-append file-name "/dzn")
                          (or (and=> (getenv "DZN_INCLUDE_PATH")
@@ -809,7 +896,7 @@ are weak-bisim equivalent"
                "lts")
         (receive (status stdout stderr)
             (observe
-             `("dzn" "graph" "--backend=lts"
+             `("dzn" ,@dzn-flags "graph" "--backend=lts"
                ,@includes
                ,@(if (not queue-size) '()
                      `("--queue-size" ,(number->string queue-size)))
